!******************************************************************************!
!                                                                              !
!    process_generation_rcl.f90                                                !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016,2017 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!

module process_generation_rcl

!------------------------------------------------------------------------------!

  use globals_rcl
  use tables_rcl, only: tensor_tables,binary_tables,qcd_rescaling_tables
  use currents_rcl, only: generate_currents
  use input_rcl, only: get_delta_uv_rcl,get_delta_ir_rcl, &
                       get_mu_uv_rcl,get_mu_ir_rcl,get_mu_ms_rcl
  use modelfile, only: is_particle_model_init_mdl,  &
                       set_particle_model_init_mdl, &
                       init_particles_mdl,          &
                       get_modelname_mdl,           &
                       get_modelgauge_mdl,          &
                       get_driver_timestamp_mdl,    &
                       print_parameters_mdl,        &
                       print_renoschemes_mdl,       &
                       print_counterterms_mdl
  use collier_interface_rcl, only: initialize_collier

!------------------------------------------------------------------------------!

  implicit none

!------------------------------------------------------------------------------!

  contains

!------------------------------------------------------------------------------!

  subroutine generate_processes_rcl

    ! This subroutine generates all defined processes.
    ! It must be called after all processes are defined. 

    real(sp) :: timeGENin,timeGENout
    real(dp) :: DeltaUV,DeltaIR,DeltaIR2,muUV,muIR,muMS
    character(len=100) :: modelname
    character(len=200) :: driver

    call get_delta_uv_rcl(DeltaUV)
    call get_delta_ir_rcl(DeltaIR,DeltaIR2)
    call get_mu_uv_rcl(muUV)
    call get_mu_ir_rcl(muIR)
    call get_mu_ms_rcl(muMS)

    if (selfenergy_offshell_setup .and. legsMax .gt. 2) then
      call error_rcl('Cannot generate processes with legs > 2' // &
                     'and Selfenergy offshell setup')
    end if

    if (.not. is_particle_model_init_mdl()) then
      call init_particles_mdl()
      call set_particle_model_init_mdl(.true.)
    end if

    call cpu_time (timeGENin)
    call binary_tables
    if (loopMax) call tensor_tables

    if (print_model_external_parameters) then
      call openOutput
      write(nx,'(1x,75("-"))')
      modelname = get_modelname_mdl()
      write(nx,'(2x,a,x,a)') "Active model:    ", adjustl(trim(modelname))
      modelname = get_modelgauge_mdl()
      write(nx,'(2x,a,x,a)') "Gauge:           ", adjustl(trim(modelname))
      driver = get_driver_timestamp_mdl()
      write(nx,'(2x,a,x,a)') "Model generation:", adjustl(trim(driver))
      write(nx,'(1x,75("-"))')

      write(nx,'(2x,a,g21.14,7x,a,g21.14,a)') &
      'Delta_UV   =',DeltaUV,'mu_UV =',muUV,' GeV'
      write(nx,'(2x,a,g21.14)') &
      'Delta_IR^2 =',DeltaIR2
      write(nx,'(2x,a,g21.14,7x,a,g21.14,a)') &
      'Delta_IR   =',DeltaIR,'mu_IR =',muIR,' GeV'
      write(nx,'(42x,a,g21.14,a)') &
      'mu_MS =',muMS,' GeV'
      if(print_model_internal_parameters) then
        call print_parameters_mdl(1)
      else
        call print_parameters_mdl(0)
      end if
      write(nx,*)
      call print_renoschemes_mdl
      if (use_active_qmasses) then
        write(nx,'(2x,a)') 'Quark masses in the running of alpha_s [GeV]:'
        write(nx,'(2x,a,g21.14,g21.14)') 'm_s^2 =',mq2(3)
        write(nx,'(2x,a,g21.14,g21.14)') 'm_c^2 =',mq2(4)
        write(nx,'(2x,a,g21.14,g21.14)') 'm_b^2 =',mq2(5)
        write(nx,'(2x,a,g21.14,g21.14)') 'm_t^2 =',mq2(6)
        write(nx,'(1x,75("-"))')
      end if
      write(nx,*)
    end if

    call generate_currents

    call cpu_time (timeGENout)

    timeGEN = timeGENout - timeGENin

    allocate (timeTI(prTot)); timeTI = 0e0
    allocate (timeTC(prTot)); timeTC = 0e0

    if (loopMax .and. init_collier) call initialize_collier

    if (qcd_rescaling) then
      call qcd_rescaling_tables
    end if

    if (print_model_counterterms) then
      call print_counterterms_mdl
      write(nx,*)
    end if

    processes_generated = .true.

  end subroutine generate_processes_rcl

!------------------------------------------------------------------------------!

end module process_generation_rcl

!------------------------------------------------------------------------------!
